/*************************************/
/* testSemaphoreEtPrioriteStatique.c */
/*************************************/

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sched.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/sem.h>

#define FATAL(msg) {\
  char m[128];       \
  sprintf(m, "%s in %s:%d:%s", msg, __FILE__, __LINE__, __func__); \
  perror(m); \
  abort(); \
}

/**********************************************************************/
/* Code de P et V                                                     */
/**********************************************************************/
void P(int semId){
  struct sembuf op;
  op.sem_num =  0;
  op.sem_op  = -1;
  op.sem_flg =  0;
  if (semop(semId, &op, 1) == -1){
    FATAL("semop");
  }
}

void V(int semId){
  struct sembuf op = {0, 1, 0};
  if (semop(semId, &op, 1) == -1){
    FATAL("semop");
  }
}

/**********************************************************************/
/* Code execute par les processus enfant                              */
/**********************************************************************/
void codeEnfant(int numEnfant, int semId)
{
  struct sched_param sp;

  /* Section de Code 1 (SC1) */
  sp.sched_priority = numEnfant;
  if (sched_setscheduler(0,SCHED_FIFO,&sp) < 0) {
    FATAL("sched_setscheduler");
  }

  /* Section de Code 2 (SC2) */
  sleep(numEnfant);

  /* Section de Code 3 (SC3) */
  P(semId);

  /* Section de Code 4 (SC4) */
  printf("Enfant %d s'est reveille !\n", numEnfant);

  exit(EXIT_SUCCESS);
}

/**********************************************************************/
/* Programme principal                                                */
/**********************************************************************/
int main(int argc, char* argv[])
{
  int status;
  key_t cle;
  int i;
  int semId;

  
  /* Section de Code 5 (SC5) */
  cle = ftok("./testSemaphoreEtPrioriteStatique.c", '0');
  if (cle < 0) {
    FATAL("ftok");
  }
  semId = semget(cle, 1, IPC_CREAT|0666);
  if (semId < 0) {
    FATAL("semget");
  }
  if (semctl(semId, 0, SETVAL, 0) < 0) {
    FATAL("semctl");
  }

  /* Section de Code 6 (SC6) */
  for (i=1 ; i<4 ; i++){
    switch (fork()) {
    case -1:
      FATAL("fork");
      break;
    case 0:
      codeEnfant(i, semId);
      break;
    default:
      break;
    }
  }

  /* Section de Code 7 (SC7) */
  sleep(4);

  /* Section de Code 8 (SC8) */
  for (i=1 ; i<4 ; i++){
    V(semId);
    if (wait(&status) < 0){
      FATAL("wait");
    }
  }

  /* Section de Code 9 (SC9) */
  if (semctl(semId, 0, IPC_RMID, 0) < 0){
    FATAL("semctl");
  }

  return EXIT_SUCCESS;
}
